#ifndef VEVENT_H
#define VEVENT_H


//: A structure for VERITAS single telescope data events coming form DACQ 
//: (Sp16) to harvester and then being saved in run files.

struct Event
{
  short int fTelescopeID;  // Telescope this event comes
                           // from. Telecopes are numbered 0-6.
  short int fEventLength;  // Specifies in bytes the total length
                           // of this event record.
                           
  int fEventNumber;        //Event trigger number as recieved by DACQ (Sp16)
                           // over Dot module from Trigger level 3(TL3) (SP10)
                           // This number is unique identifier (in a night) of
                           // Triggers generated by TL3. I.e. Run independent!

  int fStatus;             //Status word. Bits (0-32) set by DACQ and 
                           // Harvester.
                           // (Meaning of set bits should be documented here!)
                           // bit 0:on: This is normal real event.
                           // bit 1:on: Pulser generated Calibration event.  
 
  int fGPSTime[5];         //Gps time as read by Dacq(not value from TL3)

  int fNumberFadcSamples;  //Number of Fadc samples per Fadc channel in data 
                           // that follows. This is a constant:N_s=4*i where
                           // i is an integer. Typically it will be :
                           // kNumberFadcSamples=16.

  int fNumberFadcHits;     //Number of fadc channels that have data in this 
                           // record. Varies each event. Range is probably 3 
                           // to 499
  
  VPattern fChannelHits;  //512 bit pattern,(only first 499 used) indicating 
                           // which Fadc channels
                           // in this event produced data that follows. A set 
                           // bit indicates that for that channel
                           // data follows in the event record. Inclusion of a
                           // channel depends on noise supressing levels 
                           // and algoritums chosen for Fadc.
                           // The number of the bits set in this array should
                           // equal fNumFadcHits above.

  VPattern fChannelTriggers; //512 bit pattern(only first 499 used). Indicates
                           // which Cfd channels fired in this event.
                           // Possible(an error?) that a bit set here won't be
                           // set in fFadcPattern. Likely this is a subset of
                           // fFadcPattern. More channels will be read out than
                           // fire the Cfd's

  VPattern  fChannelHiLow;//512 bit pattern(only first 499 used). Indicates 
                           // which Fadc channels are read out in Hi(bit set)
                           // scale(Problem here: 1 bit per 16 samples or 
                           // should this be a [16,16] array to give 1 bit per 
                           // sample?
  
  VFadcData fFadcData[kNumberPMTs];
                            // An array of ChannelData_t  (fadc data 
                           // structure) There will be fNumFadcHits elements in
                           // this array.
}

class VEvent
{
 private:
  union
  {
    Event a;
    char b[sizeof(Event)];
  }
  bool GetFadcIndex(int channel, int fadcIndex);
 public:
  VEvent(){};
  ~VEvent(){};
  int  GetEventLength(){return a.fEventLength;};
  int  GetTelescopeID(){return1 a.TelescopeID;};
  int  GetEventNumber(){return a.fEventNumber;};
  bool GetStatus(int FlagNumber);
  void GetGPSTime(int* timeGPS);
  int  GetNumberFadcSamples(){return a.fNumberFadcSamples;};
  bool IsChannelHit(int channel){return a.fChannelHits.IsBitOn(channel);};
  bool IsChannelTrigger(int channel){return 
					a.fChannelTriggers.IsBitOn(channel););
  bool IsChannelHiLow(int channel){return a.fChannelHiLow.IsBitOn(channel););
  bool GetFadcSamples(int channel, int* samples);
  bool GetFadcPedestal(int channel, int pedestal);
  bool GetFadcCharge(int channel, int charge);
  
  void PutEventLength(int length){a.fEventLength=length;return;};
  void PutTelescope(int telescopeId){a.fTelescopeId=
				       short int(telescopeId);return;};
  void PutEventNumber(int eventNumber){a.fEventNumber=eventNumber;return;};
  void SetStatus(bool flagValue,int flagNumber);
  void PutGPSTime(int* timeGPS);
  void PutNumberFadcSamples(int numberFadcSamples){a.fNumberFadcSamples=
						    numberFadcSamples;return;);
  void SetChannelHit(int channel){a.ChannelHits.SetBit(channel);return;};
  void SetChannelTrigger(int channel){a.ChannelTriggers.SetBit(channel);
                                         return;};
  void SetChannelHiLow(int channel){a.ChannelHiLow.SetBit(channel);return;};
  void ClearChannelHit(int channel){a.ChannelHits.ClearBit(channel);return;};
  void ClearChannelTrigger(int channel){a.ChannelTriggers.ClearBit(channel);
                                           return;};
  void ClearChannelHiLow(int channel){a.ChannelHiLow.ClearBit(channel);
                                        return;};
  void PutFadcSamples(int fadcIndex, int* samples)
                           {a.fFadcData[fadcIndex].PutFadc(samples);return;};
  void PutFadcPedestal(int fadcIndex, int pedestal)
                      {a.fFadcData[fadcIndex].PutPedestal(pedestal);return;};
  void PutFadcCharge(int fadcIndex, int charge)
                          {a.fFadcData[fadcIndex].PutCharge(charge);return;};
  int GetChannelId(){return fChannelID;};
  int GetPedestal(){return fPedestal;};
  int GetCharge(){return fCharge;};
  void GetFadc(int* sample);

  void SetChannelId(int channel){fChannelId=channel;return;};
  void SetPedestal(int pedestal){fPedestal=short int pedestal;return;};
  void SetCharge(int charge){fCharge=short int charge;return;};
  void SetFadc(int* samples);
};


class VPattern
{
 private:
  int fPattern[int(kNumberPMTs/(sizeof(int)*8))+1];
 public:
  Vpattern(){};
  ~Vpattern(){};
  bool IsBitOn(int channel);
  void SetBit(int channel);
  void ClearBit(int channel);
};

class VFadcData
{
 private:
  int fChannelID;
  unsigned char  fFadcSamples[ksamples];
  short int fPedestal;
  short int fCharge;
 public:
  VFadcData(){};
  ~VFadcData(){};
  int GetChannelId(){return fChannelID;};
  int GetPedestal(){return fPedestal;};
  int GetCharge(){return fCharge;};
  void GetFadc(int* sample);

  void SetChannelId(int channel){fChannelId=channel;return;};
  void SetPedestal(int pedestal){fPedestal=short int pedestal;return;};
  void SetCharge(int charge){fCharge=short int charge;return;};
  void SetFadc(int* samples);
};

#endif //VEVENT_H

