!**************************************************************************
!	This is a set of routines for implimenting FIFO queues. It is taken
!	almost verbatim from "Upgrading to Fortran 90" by Cooper Redwine, 
!	pg318-326.

! Four subroutine/functions are defined:
!  1:QUEUE_INIT: Defines a new queue. Use looks like:
!		call queue_init(q1)	!defines and inits a queue named "q1"
!  2:QUEUE_PUT_IN_REAR: Adds a node to the rear of a queue. Use looks like:
!		call queue_put_in_rear(q1,event_number) !adds a node with
!						!ievent=event_number to end of
!						!queue
!  3:QUEUE_TAKE_FROM_FRONT: Gets value of ievent from beginning node of queue
!	removes that node and makes the next node the first node. Use looks
!	like:
!		call queue_take_from_front(q1,event_number)
!  4:QUEUE_IS_EMPTY: Test whether there are any nodes in the queue. Use looks
!	like:
!		empty=queue_is_empty(q1)	!returns true if q1 is empty.
!  

!  For my use I add 3 more functions:
!  5:QUEUE_DELETE_NODE: If a node exits with event number of event_number, 
!	remove it from the queue. reconnect adjacent nodes to each other. Use 
!	looks like:
!		call queue_delete_node(q1,event_number,node_id) !Returns
!				!node_id = node to which event was sent
!				!if a node with ievent==event_number existed 
!				!in the queue q1 and was deleted.Otherwise
!				!node_id=0.

!  6:QUEUE_EVENT_IN_FRONT:Check if the front node has been loaded with an event.
!	Use looks like:
!		event_returned=queue_event_in_front(q1)

!  7:QUEUE_SAVE_EVENT:Search through the queue for the node of event # nevent.
!	When found save the returned hntp record in the node in preperation
!	for being written to disk when its time comes. Set the event_returned
!	flag to .true. Use looks like:
!		call queue_save_event(q1,nevent,hntp)
!**************************************************************************


!	User specified node definition:
MODULE NODE_TYPE_DEF
	use structures
	IMPLICIT NONE

	type :: node_type
		integer*4 :: ievent			!event number
		logical :: event_returned		!Flag that an event for
							!this event # has been
							!written into the node.
		type (h_ntp) :: hntp			!Returned hillas record
		type (node_type),pointer :: next	!ptr to successor
							!node in queue
	end type node_type
END MODULE NODE_TYPE_DEF

!	The queue definintion module:
MODULE QUEUE_TYPE_DEF
	use node_type_def	!access definition of node type
	IMPLICIT NONE

	type queue_type
		type(node_type),pointer :: front	!ptr to first node
      		type(node_type),pointer :: rear		!ptr to last  node
       end type queue_type
END MODULE QUEUE_TYPE_DEF



!QUEUE  FUNTIONS FOLLOW
MODULE QUEUE_SUBS

CONTAINS

    SUBROUTINE QUEUE_INIT(q)
!**************************************************************************
!	Init q
!**************************************************************************
	use queue_type_def
	IMPLICIT NONE
	
	type (queue_type),intent(out) :: q
	
	nullify (q%front, q%rear)	!Disassociates pointer to front and 
					!rear of queue
    END SUBROUTINE QUEUE_INIT
!**************************************************************************

    SUBROUTINE QUEUE_PUT_IN_REAR(q,nevent)
!**************************************************************************
!	Add a node with data NEVENT to end of queue Q
!**************************************************************************
	use queue_type_def
	IMPLICIT NONE
	
	type (queue_type),intent(in out) :: q
	integer*4,intent(in) :: nevent	  

	if(.not. associated(q%front)) then	!Test is this is first entry.
		allocate(q%front)		!Make first node of queue
		q%rear => q%front		!Also the last.

	else					!Add node to end
		allocate(q%rear%next)		!New last node. Clever here.
		q%rear=>q%rear%next		!Its now the last.
	endif

	q%rear%ievent=nevent			!Save event number.
	q%rear%event_returned=.false.		!Set flag to no event yet.

	nullify(q%rear%next)			!Make sure next pointer is 
						!disasociated. Its our flag
						!that it is indeed the last.
!debug
!      print*,' put_in_rear;rear:nevent,front:nevent',q%rear%ievent, &
!           & q%front%ievent
!enddebug
    END SUBROUTINE QUEUE_PUT_IN_REAR
!**************************************************************************

    SUBROUTINE QUEUE_TAKE_FROM_FRONT(q,hntp)
!**************************************************************************
!	Get the data stored in the first node stored in the queue and remove
!	that node from the queue.
!	ASSUME HERE THAT AT LEAST 1 NODE IS IN QUEUE!
!	ASSUME HERE THAT THIS NODE HAS A RETURNED EVENT.(can be checked by
!	use ofr  QUEUE_EVENT_IN_FRONT
!**************************************************************************
	use queue_type_def
	use structures
	IMPLICIT NONE
	
	type (queue_type),intent(in out) :: q
	type (h_ntp) :: hntp
	
	type (node_type),pointer ::temp		!Place to temporarily save
						!address of successor to 
						!first node.

	hntp=q%front%hntp
	temp =>q%front%next			!Point to next(which may not
						!exist!
	deallocate(q%front)			!release node storage
	q%front=>temp				!Advace front of queue

!debug
!      print*,'Take_from_front;front%ievent:',q%front%ievent
!enddebug
    END SUBROUTINE QUEUE_TAKE_FROM_FRONT
!**************************************************************************

    FUNCTION QUEUE_EVENT_IN_FRONT(q) RESULT (event_ready)
!**************************************************************************
!	Test whether front node has had an event returned to it.
!**************************************************************************
	use queue_type_def
	IMPLICIT NONE

	logical :: event_ready		!Result variable
	type (queue_type), intent(in) :: q

	event_ready=q%front%event_returned  !Flag that front node has data
					    !ready to be written out.
!debug
!      print*,'event_in_front;front%ivent,front%event_returned:', &
!           & q%front%ievent,q%front%event_returned
!enddebug
    END FUNCTION QUEUE_EVENT_IN_FRONT
!**************************************************************************

    FUNCTION QUEUE_IS_EMPTY(q) RESULT (empty)
!**************************************************************************
!	Test whether queue has any nodes in it.
!**************************************************************************
	use queue_type_def
	IMPLICIT NONE

	logical :: empty		!Result variable
	type (queue_type), intent(in) :: q

	empty=.not.associated(q%front)	!Return .true. if q%front is
					!disassociated.
!debug
!      print*,'is_empty;empty:',empty
!enddebug
    END FUNCTION QUEUE_IS_EMPTY
!**************************************************************************

    SUBROUTINE QUEUE_DELETE_NODE(q,nevent,error)
!**************************************************************************
!	Search through the queue looking for the node that has an ievent of
!	nevent. When found delete that node.
!	If requested event number is not found in queue return error=.true.
!**************************************************************************

!	Written by G.Sembroski 03/11/99
!	Modified:

	use queue_type_def
	IMPLICIT NONE
	
	type (queue_type),intent(in out) :: q
	integer*4,intent(in) :: nevent	  
	logical :: error
	  
	type (node_type),pointer :: present	 !Node we are testing.
	type (node_type),pointer :: previous	 !Node we may repoint.

 	error=.false.			!Setup for success.
	if(queue_is_empty(q))then
		error=.true.
!debug
!	print*,'delete node;error-queue empty:nevent',nevent
!enddebug
	return
	endif
	
	present=>q%front	!Start at the front
	if(present%ievent==nevent)then   	!Its special.
		q%front=>present%next
		deallocate(present)
!debug
!	print*,'delete node;success:nevent at front',nevent
!enddebug
		return
	else
		previous=>q%front
		present=>q%front%next
	endif
	do
	   if(present%ievent==nevent)then 
		previous%next=>present%next  !skip this node in queue
						!if last it may be
						!disassociated
		deallocate(present)
!debug
!	print*,'delete node;success:nevent',nevent
!enddebug
		return
           elseif(present%ievent>nevent)then 
                                    	!Envent number increas always.
					!This is an efficeny check to save time
		error=.true.
!debug
!	print*,'delete node;error-event not found:nevent',nevent
!enddebug
		return

	   else                        !See if were at end of queue
		previous =>present
		present=>present%next  !Move to next node
		if(.not.associated(present))then	!leave if done
			error=.true.
!debug
!	print*,'delete node;error-event not found:nevent',nevent
!enddebug
			return
		endif
	   endif
	enddo
    END SUBROUTINE QUEUE_DELETE_NODE
!**************************************************************************

    SUBROUTINE QUEUE_SAVE_EVENT(q,nevent,hntp,error)
!**************************************************************************
!	Search through the queue looking for the node that has an ievent of
!	nevent. When found save the hntp into that node.
!**************************************************************************

!	Written by G.Sembroski 03/11/99
!	Modified:

	use queue_type_def
	use structures
	IMPLICIT NONE
	
	type (queue_type),intent(in out) :: q
	integer*4,intent(in) :: nevent	  
	type(h_ntp) :: hntp
	logical :: error
	  
	type (node_type),pointer :: present	 !Node we are testing.

	error=.false.			!Preset flag to success
	if(queue_is_empty(q))then
		error=.true.
!debug
!	print*,'save_event;error-queue empty:nevent',nevent
!enddebug
		return
	endif
	
	present=>q%front	!Start at the front
	do
		if(present%ievent==nevent)then   	!Found it
			present%hntp=hntp
			present%event_returned=.true.   !Flag it there.
!debug
!	print*,'save_event;success:nevent',nevent
!enddebug
			return
		elseif(present%ievent>nevent)then       !Its not there
			error=.true.
!debug
!	print*,'save_event;error-nevent not found:nevent',nevent
!enddebug
			return
		else
	 		present=>present%next
			if(.not.associated(present))then	!leave if done
				error=.true.
!debug
!	print*,'save_event;error-nevent not found:nevent',nevent
!enddebug
				return
			endif
		endif
	enddo
    END SUBROUTINE QUEUE_SAVE_EVENT
!**************************************************************************

END MODULE QUEUE_SUBS

!	Queue_interface:
MODULE QUEUE_INTERFACE
	use queue_type_def
	use queue_subs
	IMPLICIT NONE

END MODULE QUEUE_INTERFACE


	

